@@ -1,6 +1,6 @@
 // (C) Copyright International Business Machines Corporation 2007, 2008
 // All Rights Reserved.
-// This code is published under the Common Public License.
+// This code is published under the Eclipse Public License.
 //
 // Authors :
 // Andreas Waechter, International Business Machines Corporation
@@ -14,12 +14,14 @@
 #include "BonBqpdWarmStart.hpp"
 
 #include "CoinTime.hpp"
+#include <algorithm>
 
 #define InitializeAll
 
 typedef Bonmin::BqpdSolver::fint fint;
 typedef Bonmin::BqpdSolver::real real;
-
+int Bonmin::BqpdSolver::reinit_freq_ = 0;
+int Bonmin::BqpdSolver::m0de_ = 6;
 extern "C"
 {
   void F77_FUNC(bqpd,BQPD)(fint* n, fint* m, fint* k, fint* kmax,
@@ -108,6 +110,8 @@ namespace Bonmin
   {
     roptions->SetRegisteringCategory("Bqpd options",RegisteredOptions::BqpdCategory);
     roptions->AddLowerBoundedNumberOption("qp_fillin_factor", "Factor for estimating fill-in for factorization method in Bqpd", 0., true, 4.);
+    roptions->AddBoundedIntegerOption("hot_start_m0de", "Choice of the hot start option", 0, 6, 6);
+    roptions->AddLowerBoundedIntegerOption("reinit_freq", "Number of pivots before recopy hot start", 0, 0);
   }
 
   BqpdSolver::BqpdSolver(bool createEmpty /* = false */)
@@ -120,26 +124,42 @@ namespace Bonmin
 
   BqpdSolver::BqpdSolver(Ipopt::SmartPtr<Bonmin::RegisteredOptions> roptions,
       Ipopt::SmartPtr<Ipopt::OptionsList> options,
-      Ipopt::SmartPtr<Ipopt::Journalist> journalist)
+      Ipopt::SmartPtr<Ipopt::Journalist> journalist,
+      const std::string & prefix)
       :
-      TNLPSolver(roptions, options, journalist),
+      TNLPSolver(roptions, options, journalist, prefix),
       cached_(NULL)
   {
     options->GetNumericValue("qp_fillin_factor", fillin_factor_, "bqpd.");
     options->GetIntegerValue("kmax", kmax_ipt_, "bqpd.");
     options->GetIntegerValue("mlp", mlp_ipt_,"bqpd.");
+    options->GetIntegerValue("hot_start_m0de", m0de_,"bqpd.");
+    options->GetIntegerValue("reinit_freq", reinit_freq_,"bqpd.");
+     if(!options_->GetIntegerValue("print_level",default_log_level_,""))
+      default_log_level_ = 1;
+
   }
 
   Ipopt::SmartPtr <TNLPSolver>
   BqpdSolver::clone()
   {
+#ifdef TIME_BQPD
+    times().create -= CoinCpuTime();
+#endif
     Ipopt::SmartPtr<BqpdSolver> retval = new BqpdSolver(true);
     *retval->options_ = *options_; // Copy the options
     retval->roptions_ = roptions_; // only copy pointers of registered options
     retval->journalist_ = journalist_; // and journalist
+    retval->prefix_ = prefix_;
     retval->fillin_factor_ = fillin_factor_;
     retval->kmax_ipt_ = kmax_ipt_;
     retval->mlp_ipt_ = mlp_ipt_;
+    retval->default_log_level_ = default_log_level_;
+    retval->reinit_freq_ = reinit_freq_;
+    retval->m0de_ = m0de_;
+#ifdef TIME_BQPD
+    times().create += CoinCpuTime();
+#endif
     return GetRawPtr(retval);
   }
 
@@ -168,6 +188,8 @@ namespace Bonmin
     if (is) {
       is.close();
     }
+    if(!options_->GetIntegerValue("print_level",default_log_level_,""))
+      default_log_level_ = 1;
     return retval;
   }
 
@@ -194,7 +216,14 @@ namespace Bonmin
       cached_ = new cachedInfo(tqp, options_, kmax_ipt_, mlp_ipt_,
 			       &fillin_factor_);
     }
-    return callOptimizer();
+#ifdef TIME_BQPD
+    times().solve -= CoinCpuTime();
+#endif
+    TNLPSolver::ReturnStatus r_val = callOptimizer();
+#ifdef TIME_BQPD
+    times().solve += CoinCpuTime();
+#endif
+    return r_val;
   }
 
 /// Solves a problem expresses as a TNLP
@@ -211,12 +240,26 @@ namespace Bonmin
         m, cached_->bl+n, cached_->bu+n);
     // Make sure bounds are not infinity
     for (int i=0; i<n+m; i++) {
-      cached_->bl[i] = Max(cached_->bl[i], -1e50);
-      cached_->bu[i] = Min(cached_->bu[i], 1e50);
+      cached_->bl[i] = std::max(cached_->bl[i], -1e50);
+      cached_->bu[i] = std::min(cached_->bu[i], 1e50);
     }
 
+#if 1
     cached_->use_warm_start_in_cache_ = true;  // Trying...
-    return callOptimizer();
+    cached_->m0de = m0de_;
+#else
+    cached_->re_initialize();
+    cached_->use_warm_start_in_cache_ = true;  // Trying...
+#endif
+
+#ifdef TIME_BQPD
+    times().resolve -= CoinCpuTime();
+#endif
+    TNLPSolver::ReturnStatus r_val = callOptimizer();
+#ifdef TIME_BQPD
+    times().resolve += CoinCpuTime();
+#endif
+    return r_val;
   }
 
   void
@@ -241,7 +284,7 @@ namespace Bonmin
     F77_FUNC(refactorc,REFACTORC).nfreq = 500;
 
     Ipopt::TNLP::IndexStyleEnum index_style;
-    Index nv, nc, nnz_jac_g, nnz_hess;
+    Ipopt::Index nv, nc, nnz_jac_g, nnz_hess;
     tqp->get_nlp_info(nv, nc, nnz_jac_g, nnz_hess, index_style);
     n = nv;
     m = nc;
@@ -251,7 +294,7 @@ namespace Bonmin
     }
     else {
       kmax = kmax_ipt;
-      kmax = min(kmax,n);
+      kmax = std::min(kmax,n);
     }
     mlp = mlp_ipt;
 
@@ -274,13 +317,13 @@ namespace Bonmin
 
     // Make sure bounds are not infinity
     for (int i=0; i<n+m; i++) {
-      bl[i] = Max(bl[i], -1e50);
-      bu[i] = Min(bu[i], 1e50);
+      bl[i] = std::max(bl[i], -1e50);
+      bu[i] = std::min(bu[i], 1e50);
     }
 
     // Set up sparse matrix with objective gradient and constraint Jacobian
 
-    const Number* obj_grad = tqp->ObjGrad();
+    const Ipopt::Number* obj_grad = tqp->ObjGrad();
     amax_ = nnz_jac_g;
     for (int i = 0; i<n; i++) {
       if (obj_grad[i]!=0.) {
@@ -302,13 +345,13 @@ namespace Bonmin
     la[amax_+1] = 1;
 
     // Constraint Jacobian
-    const Number* JacVals = tqp->ConstrJacVals();
-    const Index* RowJac = tqp->ConstrJacIRow();
-    const Index* ColJac = tqp->ConstrJacJCol();
+    const Ipopt::Number* JacVals = tqp->ConstrJacVals();
+    const Ipopt::Index* RowJac = tqp->ConstrJacIRow();
+    const Ipopt::Index* ColJac = tqp->ConstrJacJCol();
 
     int* permutationJac = new int [nnz_jac_g];
     FilterSolver::TMat2RowPMat(false, n, m, nnz_jac_g,  RowJac, ColJac, permutationJac,
-        la, nnz_grad, 1, TNLP::C_STYLE);
+        la, nnz_grad, 1, Ipopt::TNLP::C_STYLE);
     for (int i=0; i<nnz_jac_g; i++) {
       const int& indice = permutationJac[i];
       a[nnz_grad+i] = JacVals[indice];
@@ -338,16 +381,16 @@ namespace Bonmin
 #endif
 
     // Now setup Hessian
-    const Number* HessVals = tqp->ObjHessVals();
-    const Index* RowHess = tqp->ObjHessIRow();
-    const Index* ColHess = tqp->ObjHessJCol();
+    const Ipopt::Number* HessVals = tqp->ObjHessVals();
+    const Ipopt::Index* RowHess = tqp->ObjHessIRow();
+    const Ipopt::Index* ColHess = tqp->ObjHessJCol();
 
     kk = nnz_hess;
     ll = nnz_hess + n + 2;
     int* permutationHess = new int[nnz_hess];
 
     FilterSolver::TMat2RowPMat(true, n, n, nnz_hess, RowHess, ColHess,
-        permutationHess, lws, 0, 0, TNLP::C_STYLE);
+        permutationHess, lws, 0, 0, Ipopt::TNLP::C_STYLE);
     for (int i=0; i<nnz_hess; i++) {
       ws[i] = HessVals[permutationHess[i]];
     }
@@ -359,7 +402,7 @@ namespace Bonmin
     for (int i=0; i<kk; i++) printf("hess ws[%3d] = %e\n",i,ws[i]);
 #endif
 
-    Index bufy;
+    Ipopt::Index bufy;
     options->GetIntegerValue("iprint",bufy, "bqpd.");
     iprint = bufy;
     nout = 6;
@@ -367,10 +410,25 @@ namespace Bonmin
     tqp_ = tqp;
   }
 
-/// Solves a problem expresses as a TNLP
-  TNLPSolver::ReturnStatus
-  BqpdSolver::callOptimizer()
+  void
+  BqpdSolver::cachedInfo::re_initialize()
   {
+    use_warm_start_in_cache_ = false;
+
+    // Setup workspaces
+    for (int i=0; i<mxws; i++) {
+      ws[i] = 42.;
+    }
+    for (int i=0; i<mxlws; i++) {
+      lws[i] = 55;
+    }
+  }
+
+/// Solves a problem expresses as a TNLP
+  TNLPSolver::ReturnStatus BqpdSolver::callOptimizer() {
+#ifdef TIME_BQPD
+    times().numsolve++;
+#endif
     cached_->optimize();
 
     TNLPSolver::ReturnStatus optimizationStatus = TNLPSolver::exception;
@@ -391,8 +449,8 @@ namespace Bonmin
       break;
     }
 
-    Index dummy_len = Ipopt::Max(cached_->n,cached_->m);
-    Number* dummy = new Number[dummy_len];
+    Ipopt::Index dummy_len = std::max(cached_->n,cached_->m);
+    Ipopt::Number* dummy = new Ipopt::Number[dummy_len];
     for (int i=0; i<dummy_len; i++) {
       dummy[i] = 0.;
     }
@@ -429,6 +487,13 @@ namespace Bonmin
   bool
   BqpdSolver::cachedInfo::markHotStart()
   {
+     next_reinit_ = BqpdSolver::reinit_freq_; 
+#ifdef DISABLE_COPYING
+     return 1;
+#endif
+#ifdef TIME_BQPD
+    times_.warm_start -= CoinCpuTime();
+#endif
     haveHotStart_ = true;
     irh1 = F77_FUNC(bqpdc,BQPDC).irh1;
     na = F77_FUNC(bqpdc,BQPDC).na;
@@ -492,7 +557,10 @@ namespace Bonmin
 
     mxm1 = F77_FUNC(mxm1c,MXM1c).mxm1;
     c = F77_FUNC(minorc,MINORC).c;
+
+    kkHot = F77_FUNC(wsc,WSC).kk;
     kkkHot = F77_FUNC(wsc,WSC).kkk;
+    llHot = F77_FUNC(wsc,WSC).ll;
     lllHot = F77_FUNC(wsc,WSC).lll;
 
     kHot = k;
@@ -509,6 +577,9 @@ namespace Bonmin
     wsHot = CoinCopyOfArray(ws,mxws);
     lwsHot = CoinCopyOfArray(lws,mxlws);
     infoHot[0] = info[0];
+#ifdef TIME_BQPD
+    times_.warm_start += CoinCpuTime();
+#endif
 
     return true;
   }
@@ -516,6 +587,12 @@ namespace Bonmin
   void
   BqpdSolver::cachedInfo::copyFromHotStart()
   {
+#ifdef DISABLE_COPYING
+    return;
+#endif
+#ifdef TIME_BQPD
+    times_.warm_start -= CoinCpuTime();
+#endif
     F77_FUNC(bqpdc,BQPDC).irh1 = irh1;
     F77_FUNC(bqpdc,BQPDC).na = na;
     F77_FUNC(bqpdc,BQPDC).na1 = na1;
@@ -579,7 +656,9 @@ namespace Bonmin
     F77_FUNC(mxm1c,MXM1c).mxm1 = mxm1;
     F77_FUNC(minorc,MINORC).c = c;
 
+    F77_FUNC(wsc,WSC).kk = kkHot;
     F77_FUNC(wsc,WSC).kkk = kkkHot;
+    F77_FUNC(wsc,WSC).ll = llHot;
     F77_FUNC(wsc,WSC).lll = lllHot;
 
     k = kHot;
@@ -592,10 +671,14 @@ namespace Bonmin
     CoinCopyN(lsHot,n+m,ls);
     CoinCopyN(alpHot,mlp,alp);
     CoinCopyN(lpHot,mlp,lp);
-    peq = peqHot;
     CoinCopyN(wsHot,mxws,ws);
     CoinCopyN(lwsHot,mxlws,lws);
     info[0] = infoHot[0];
+    peq = peqHot;
+#ifdef TIME_BQPD
+    times_.warm_start += CoinCpuTime();
+#endif
+
   }
 
   void
@@ -628,10 +711,14 @@ namespace Bonmin
     F77_FUNC(wsc,WSC).mxlws = mxlws;
 
     if (use_warm_start_in_cache_ && !bad_warm_start_info_) {
-      m0de = 6;
       ifail = 0;
       use_warm_start_in_cache_ = false;
-      if (haveHotStart_) copyFromHotStart();
+      if (haveHotStart_ && pivots_ > next_reinit_) {
+        //printf("Reinitialize hot start\n");
+        copyFromHotStart();
+        while (BqpdSolver::reinit_freq_ > 0&& next_reinit_ < pivots_)
+          next_reinit_ += BqpdSolver::reinit_freq_;
+      }
     }
     else {
       m0de = 0;
@@ -656,12 +743,68 @@ namespace Bonmin
     for (int i=0; i<n; i++) {
       printf("qxstart[%2d] = %23.16e\n", i, x[i]);
     }
+#endif
+//#define WRITE_QPS
+#ifdef WRITE_QPS
+    if (m0de==0) {
+      FILE* fp = fopen("QPinit.dat", "w");
+      fprintf(fp, "n = %d\n", n);
+      fprintf(fp, "m = %d\n", m);
+      fprintf(fp, "kmax = %d\n", kmax);
+      fprintf(fp, "amax = %d\n" ,amax_);
+      for (int i=1; i<=amax_; i++) {
+        fprintf(fp, "a = %23.16e\n", a[i-1]);
+      }
+      int lamax = amax_ + m + 2;
+      fprintf(fp, "lamax = %d\n" ,lamax);
+      for (int i=1; i<=lamax+1; i++) {
+        fprintf(fp, "la = %6d\n", la[i-1]);
+      }
+      for (int i=1; i<=n; i++) {
+        fprintf(fp, "x = %23.16e\n", x[i-1]);
+      }
+      for (int i=1; i<=n+m; i++) {
+        fprintf(fp, "bl = %23.16e\n", bl[i-1]);
+      }
+      for (int i=1; i<=n+m; i++) {
+        fprintf(fp, "bu = %23.16e\n", bu[i-1]);
+      }
+      fprintf(fp, "fmin = %23.16e\n", fmin);
+      fprintf(fp, "mlp = %6d\n", mlp);
+      fprintf(fp, "mxws = %d\n", mxws);
+      fprintf(fp, "mxlws = %d\n", mxlws);
+      fclose(fp);
+    }
+    else {
+      FILE* fp = fopen("QPbounds.dat", "w");
+       fprintf(fp, "m0de = %d\n", m0de);
+      for (int i=1; i<=n+m; i++) {
+        fprintf(fp, "bl = %23.16e\n", bl[i-1]);
+      }
+      for (int i=1; i<=n+m; i++) {
+        fprintf(fp, "bu = %23.16e\n", bu[i-1]);
+      }
+      fclose(fp);
+    }
 #endif
     F77_FUNC(bqpd,BQPD)(&n, &m, &k, &kmax, a, la, x, bl, bu, &f, &fmin,
         g, r, w, e, ls, alp, lp, &mlp, &peq, ws, lws,
         &m0de, &ifail, info, &iprint, &nout);
-    if (ifail == 8 && m0de == 6) {
-      printf("Restarting Bqpd...");
+#ifdef TIME_BQPD
+    times_.pivots += info[0];
+#endif
+    pivots_ += info[0];
+    if(BqpdSolver::reinit_freq_ > 0 && haveHotStart_ && (ifail == 7 || ifail == 8) && m0de == 6){
+      fprintf(stdout, "Reinitialize hot start...\n");
+      copyFromHotStart();
+      ifail = 0;
+      F77_FUNC(bqpd,BQPD)(&n, &m, &k, &kmax, a, la, x, bl, bu, &f, &fmin,
+			  g, r, w, e, ls, alp, lp, &mlp, &peq, ws, lws,
+			  &m0de, &ifail, info, &iprint, &nout);
+      printf("new ifail = %d\n", ifail);
+    }
+    if (ifail == 8 && BqpdSolver::m0de_ == 6) {
+      fprintf(stdout, "Restarting Bqpd...");
       m0de = 0;
       tqp_->get_starting_point(n, 1, x, 0, NULL, NULL, m, 0, NULL);
       ifail = 0;
