@@ -1,8 +1,8 @@
-// Copyright (C) 2004, 2009 International Business Machines and others.
+// Copyright (C) 2004, 2011 International Business Machines and others.
 // All Rights Reserved.
-// This code is published under the Common Public License.
+// This code is published under the Eclipse Public License.
 //
-// $Id: IpAlgBuilder.cpp 1578 2009-10-06 23:40:31Z andreasw $
+// $Id: IpAlgBuilder.cpp 2338 2013-06-16 21:16:20Z stefan $
 //
 // Authors:  Carl Laird, Andreas Waechter     IBM    2004-09-29
 
@@ -40,6 +40,7 @@
 #include "IpLimMemQuasiNewtonUpdater.hpp"
 #include "IpOrigIpoptNLP.hpp"
 #include "IpLowRankAugSystemSolver.hpp"
+#include "IpLowRankSSAugSystemSolver.hpp"
 #include "IpRestoIterationOutput.hpp"
 #include "IpRestoFilterConvCheck.hpp"
 #include "IpRestoIterateInitializer.hpp"
@@ -51,10 +52,17 @@
 #include "IpEquilibrationScaling.hpp"
 #include "IpExactHessianUpdater.hpp"
 
+#ifdef COIN_HAS_HSL
+#include "CoinHslConfig.h"
+#endif
 #include "IpMa27TSolverInterface.hpp"
 #include "IpMa57TSolverInterface.hpp"
+#include "IpMa77SolverInterface.hpp"
+#include "IpMa86SolverInterface.hpp"
+#include "IpMa97SolverInterface.hpp"
 #include "IpMc19TSymScalingMethod.hpp"
 #include "IpPardisoSolverInterface.hpp"
+#include "IpSlackBasedTSymScalingMethod.hpp"
 
 #ifdef HAVE_WSMP
 # include "IpWsmpSolverInterface.hpp"
@@ -64,8 +72,10 @@
 # include "IpMumpsSolverInterface.hpp"
 #endif
 
-#include "HSLLoader.h"
-#include "PardisoLoader.h"
+#ifdef HAVE_LINEARSOLVERLOADER
+# include "HSLLoader.h"
+# include "PardisoLoader.h"
+#endif
 
 namespace Ipopt
 {
@@ -130,25 +140,37 @@ namespace Ipopt
   void AlgorithmBuilder::RegisterOptions(SmartPtr<RegisteredOptions> roptions)
   {
     roptions->SetRegisteringCategory("Linear Solver");
-    roptions->AddStringOption6(
+    roptions->AddStringOption9(
       "linear_solver",
       "Linear solver used for step computations.",
-#ifdef HAVE_MA27
+#ifdef COINHSL_HAS_MA27
       "ma27",
 #else
-# ifdef HAVE_MA57
+# ifdef COINHSL_HAS_MA57
       "ma57",
 # else
-#  ifdef HAVE_PARDISO
-      "pardiso",
-#  else
-#   ifdef HAVE_WSMP
-      "wsmp",
+# ifdef COINHSL_HAS_MA97
+      "ma97",
+#else
+#   ifdef COINHSL_HAS_MA86
+       "ma86",
 #   else
-#    ifdef COIN_HAS_MUMPS
-      "mumps",
+#    ifdef HAVE_PARDISO
+       "pardiso",
 #    else
-      "ma27",
+#     ifdef HAVE_WSMP
+       "wsmp",
+#     else
+#      ifdef COIN_HAS_MUMPS
+       "mumps",
+#      else
+#       ifdef COINHSL_HAS_MA77
+        "ma77",
+#       else
+        "ma27",
+#       endif
+#      endif
+#     endif
 #    endif
 #   endif
 #  endif
@@ -156,6 +178,9 @@ namespace Ipopt
 #endif
       "ma27", "use the Harwell routine MA27",
       "ma57", "use the Harwell routine MA57",
+      "ma77", "use the Harwell routine HSL_MA77",
+      "ma86", "use the Harwell routine HSL_MA86",
+      "ma97", "use the Harwell routine HSL_MA97",
       "pardiso", "use the Pardiso package",
       "wsmp", "use WSMP package",
       "mumps", "use MUMPS package",
@@ -167,21 +192,22 @@ namespace Ipopt
       "to choose. Depending on your Ipopt installation, not all options are "
       "available.");
     roptions->SetRegisteringCategory("Linear Solver");
-    roptions->AddStringOption2(
+    roptions->AddStringOption3(
       "linear_system_scaling",
       "Method for scaling the linear system.",
-#ifdef HAVE_MC19
+#ifdef COINHSL_HAS_MC19
       "mc19",
 #else
       "none",
 #endif
       "none", "no scaling will be performed",
       "mc19", "use the Harwell routine MC19",
+      "slack-based", "use the slack values",
       "Determines the method used to compute symmetric scaling "
       "factors for the augmented system (see also the "
       "\"linear_scaling_on_demand\" option).  This scaling is independent "
       "of the NLP problem scaling.  By default, MC19 is only used if MA27 or "
-      "MA57 are selected as linear solvers. This option is only available if "
+      "MA57 are selected as linear solvers. This value is only available if "
       "Ipopt has been compiled with MC19.");
 
     roptions->SetRegisteringCategory("NLP Scaling");
@@ -229,7 +255,16 @@ namespace Ipopt
       "strategy. (Only considered if \"adaptive\" is selected for option "
       "\"mu_strategy\".)");
 
-    roptions->SetRegisteringCategory("Undocumented");
+    roptions->SetRegisteringCategory("Hessian Approximation");
+    roptions->AddStringOption2(
+      "limited_memory_aug_solver",
+      "Strategy for solving the augmented system for low-rank Hessian.",
+      "sherman-morrison",
+      "sherman-morrison", "use Sherman-Morrison formula",
+      "extended", "use an extended augmented system",
+      "");
+
+    roptions->SetRegisteringCategory("Line Search");
     roptions->AddStringOption3(
       "line_search_method",
       "Globalization method used in backtracking line search",
@@ -237,7 +272,9 @@ namespace Ipopt
       "filter", "Filter method",
       "cg-penalty", "Chen-Goldfarb penalty function",
       "penalty", "Standard penalty function",
-      "");
+      "Only the \"filter\" choice is officially supported.  But sometimes, "
+      "good results might be obtained with the other choices.");
+    roptions->SetRegisteringCategory("Undocumented");
     roptions->AddStringOption2(
       "wsmp_iterative",
       "Switches to iterative solver in WSMP.",
@@ -269,18 +306,20 @@ namespace Ipopt
     options.GetStringValue("linear_solver", linear_solver, prefix);
     bool use_custom_solver = false;
     if (linear_solver=="ma27") {
-#ifndef HAVE_MA27
+#ifndef COINHSL_HAS_MA27
 # ifdef HAVE_LINEARSOLVERLOADER
       SolverInterface = new Ma27TSolverInterface();
-      char buf[256];
-      int rc = LSL_loadHSL(NULL, buf, 255);
-      if (rc) {
-        std::string errmsg;
-        errmsg = "Selected linear solver MA27 not available.\nTried to obtain MA27 from shared library \"";
-        errmsg += LSL_HSLLibraryName();
-        errmsg += "\", but the following error occured:\n";
-        errmsg += buf;
-        THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+      if (!LSL_isMA27available()) {
+        char buf[256];
+        int rc = LSL_loadHSL(NULL, buf, 255);
+        if (rc) {
+          std::string errmsg;
+          errmsg = "Selected linear solver MA27 not available.\nTried to obtain MA27 from shared library \"";
+          errmsg += LSL_HSLLibraryName();
+          errmsg += "\", but the following error occured:\n";
+          errmsg += buf;
+          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        }
       }
 # else
       THROW_EXCEPTION(OPTION_INVALID, "Support for MA27 has not been compiled into Ipopt.");
@@ -291,18 +330,20 @@ namespace Ipopt
 
     }
     else if (linear_solver=="ma57") {
-#ifndef HAVE_MA57
+#ifndef COINHSL_HAS_MA57
 # ifdef HAVE_LINEARSOLVERLOADER
       SolverInterface = new Ma57TSolverInterface();
-      char buf[256];
-      int rc = LSL_loadHSL(NULL, buf, 255);
-      if (rc) {
-        std::string errmsg;
-        errmsg = "Selected linear solver MA57 not available.\nTried to obtain MA57 from shared library \"";
-        errmsg += LSL_HSLLibraryName();
-        errmsg += "\", but the following error occured:\n";
-        errmsg += buf;
-        THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+      if (!LSL_isMA57available()) {
+        char buf[256];
+        int rc = LSL_loadHSL(NULL, buf, 255);
+        if (rc) {
+          std::string errmsg;
+          errmsg = "Selected linear solver MA57 not available.\nTried to obtain MA57 from shared library \"";
+          errmsg += LSL_HSLLibraryName();
+          errmsg += "\", but the following error occured:\n";
+          errmsg += buf;
+          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        }
       }
 # else
       THROW_EXCEPTION(OPTION_INVALID, "Support for MA57 has not been compiled into Ipopt.");
@@ -311,6 +352,54 @@ namespace Ipopt
       SolverInterface = new Ma57TSolverInterface();
 #endif
 
+    }
+    else if (linear_solver=="ma77") {
+#ifndef COINHSL_HAS_MA77
+# ifdef HAVE_LINEARSOLVERLOADER
+      SolverInterface = new Ma77SolverInterface();
+      if (!LSL_isMA77available()) {
+        char buf[256];
+        int rc = LSL_loadHSL(NULL, buf, 255);
+        if (rc) {
+          std::string errmsg;
+          errmsg = "Selected linear solver HSL_MA77 not available.\nTried to obtain HSL_MA77 from shared library \"";
+          errmsg += LSL_HSLLibraryName();
+          errmsg += "\", but the following error occured:\n";
+          errmsg += buf;
+          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        }
+      }
+# else
+      THROW_EXCEPTION(OPTION_INVALID, "Support for HSL_MA77 has not been compiled into Ipopt.");
+# endif
+#else
+      SolverInterface = new Ma77SolverInterface();
+#endif
+
+    }
+    else if (linear_solver=="ma86") {
+#ifndef COINHSL_HAS_MA86
+# ifdef HAVE_LINEARSOLVERLOADER
+      SolverInterface = new Ma86SolverInterface();
+      if (!LSL_isMA86available()) {
+        char buf[256];
+        int rc = LSL_loadHSL(NULL, buf, 255);
+        if (rc) {
+          std::string errmsg;
+          errmsg = "Selected linear solver HSL_MA86 not available.\nTried to obtain HSL_MA86 from shared library \"";
+          errmsg += LSL_HSLLibraryName();
+          errmsg += "\", but the following error occured:\n";
+          errmsg += buf;
+          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        }
+      }
+# else
+      THROW_EXCEPTION(OPTION_INVALID, "Support for HSL_MA86 has not been compiled into Ipopt.");
+# endif
+#else
+      SolverInterface = new Ma86SolverInterface();
+#endif
+
     }
     else if (linear_solver=="pardiso") {
 #ifndef HAVE_PARDISO
@@ -333,6 +422,30 @@ namespace Ipopt
       SolverInterface = new PardisoSolverInterface();
 #endif
 
+    }
+    else if (linear_solver=="ma97") {
+#ifndef COINHSL_HAS_MA97
+# ifdef HAVE_LINEARSOLVERLOADER
+      SolverInterface = new Ma97SolverInterface();
+      if (!LSL_isMA97available()) {
+        char buf[256];
+        int rc = LSL_loadHSL(NULL, buf, 255);
+        if (rc) {
+          std::string errmsg;
+          errmsg = "Selected linear solver HSL_MA97 not available.\nTried to obtain HSL_MA97 from shared library \"";
+          errmsg += LSL_HSLLibraryName();
+          errmsg += "\", but the following error occured:\n";
+          errmsg += buf;
+          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        }
+      }
+# else
+      THROW_EXCEPTION(OPTION_INVALID, "Support for HSL_MA97 has not been compiled into Ipopt.");
+# endif
+#else
+      SolverInterface = new Ma97SolverInterface();
+#endif
+
     }
     else if (linear_solver=="wsmp") {
 #ifdef HAVE_WSMP
@@ -376,22 +489,24 @@ namespace Ipopt
       std::string linear_system_scaling;
       if (!options.GetStringValue("linear_system_scaling",
                                   linear_system_scaling, prefix)) {
-        // By default, don't use mc19 for non-HSL solvers
-        if (linear_solver!="ma27" && linear_solver!="ma57") {
+        // By default, don't use mc19 for non-HSL solvers, or HSL_MA97
+        if (linear_solver!="ma27" && linear_solver!="ma57" && linear_solver!="ma77" && linear_solver!="ma86") {
           linear_system_scaling="none";
         }
       }
       if (linear_system_scaling=="mc19") {
-#ifndef HAVE_MC19
+#ifndef COINHSL_HAS_MC19
 # ifdef HAVE_LINEARSOLVERLOADER
         ScalingMethod = new Mc19TSymScalingMethod();
-        char buf[256];
-        int rc = LSL_loadHSL(NULL, buf, 255);
-        if (rc) {
-          std::string errmsg;
-          errmsg = "Selected linear system scaling method MC19 not available.\n";
-          errmsg += buf;
-          THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+        if (!LSL_isMC19available()) {
+          char buf[256];
+          int rc = LSL_loadHSL(NULL, buf, 255);
+          if (rc) {
+            std::string errmsg;
+            errmsg = "Selected linear system scaling method MC19 not available.\n";
+            errmsg += buf;
+            THROW_EXCEPTION(OPTION_INVALID, errmsg.c_str());
+          }
         }
 # else
         THROW_EXCEPTION(OPTION_INVALID, "Support for MC19 has not been compiled into Ipopt.");
@@ -401,6 +516,9 @@ namespace Ipopt
 #endif
 
       }
+      else if (linear_system_scaling=="slack-based") {
+        ScalingMethod = new SlackBasedTSymScalingMethod();
+      }
 
       SmartPtr<SymLinearSolver> ScaledSolver =
         new TSymLinearSolver(SolverInterface, ScalingMethod);
@@ -413,9 +531,33 @@ namespace Ipopt
     HessianApproximationType hessian_approximation =
       HessianApproximationType(enum_int);
     if (hessian_approximation==LIMITED_MEMORY) {
-      SmartPtr<AugSystemSolver> tmp =
-        new LowRankAugSystemSolver(*AugSolver);
-      AugSolver = tmp;
+      std::string lm_aug_solver;
+      options.GetStringValue("limited_memory_aug_solver", lm_aug_solver,
+                             prefix);
+      if (lm_aug_solver == "sherman-morrison") {
+        AugSolver = new LowRankAugSystemSolver(*AugSolver);
+      }
+      else if (lm_aug_solver == "extended") {
+        Index lm_history;
+        options.GetIntegerValue("limited_memory_max_history", lm_history,
+                                prefix);
+        Index max_rank;
+        std::string lm_type;
+        options.GetStringValue("limited_memory_update_type", lm_type, prefix);
+        if (lm_type == "bfgs") {
+          max_rank = 2*lm_history;
+        }
+        else if (lm_type == "sr1") {
+          max_rank = lm_history;
+        }
+        else {
+          THROW_EXCEPTION(OPTION_INVALID, "Unknown value for option \"limited_memory_update_type\".");
+        }
+        AugSolver = new LowRankSSAugSystemSolver(*AugSolver, max_rank);
+      }
+      else {
+        THROW_EXCEPTION(OPTION_INVALID, "Unknown value for option \"limited_memory_aug_solver\".");
+      }
     }
 
     SmartPtr<PDPerturbationHandler> pertHandler;
